### How to become a Data Scientist
The Data Scientist is a multi-diciplinary job, therefore we'll explore skills of DataScientist here on example of Kaggle [Homesite](https://www.kaggle.com/c/homesite-quote-conversion) competition. Here is its description

> Using an anonymized database of information on customer and sales activity,
> including property and coverage information, Homesite is challenging you to
> predict which customers will purchase a given quote. Accurately predicting
> conversion would help Homesite better understand the impact of proposed
> pricing changes and maintain an ideal portfolio of customer segments.

### Setting up the environment
Every DataScientists needs lots of tools at his/her disposal. The easiest way to get them is to use your favoir package management system. On Linux (and UNIX) flavours we can use [Anaconda](https://www.anaconda.com/download/#macos) tool to download and install for us these tools.

Once installed you need to source its environment and be able to install any packages:

```
# setup conda
source /opt/anaconda3/etc/profile.d/conda.sh
# you can add additional PYTHON path (where your packages may reside) as following
export PYTHONPATH=$PYTHONPATH:<YOUR_PATH_HERE>

# get information about your anaconda install
conda info
conda update conda
conda env list

# create conda environment
conda create --name myenv python=3.6

# example how to install packages via conda tool
conda install -c conda-forge xgboost
conda install -c conda-forge sklearn
conda install -c conda-forge scikit-learn
conda install -c conda-forge matplotlib
conda install -c conda-forge freetype
conda install -c conda-forge keras

# activate required environment
conda activate myenv

# due to conflict in libs here is a valid setting
export KMP_DUPLICATE_LIB_OK=TRUE
```

### Review our dataset
The Homesite [dataset](https://www.kaggle.com/c/homesite-quote-conversion/data) is quite large and consist of the following files:

- train dataset, 200MB
- test dataset, 130MB
- example of submission file

Once downloaded we can start playing with these data. For that we'll use [Python](https://www.python.org/downloads/) and [R-language](https://www.r-project.org/about.html) to explore our data.

Python session:
```
# load python shell
$ python

# load data into python
>>> import pandas as pd
>>> df = pd.read_csv("data/train.csv")
>>> df.head()
>>> df.describe()

# notice how much memory our python process consume
top -p <python pid>
```

R session:
```
# load R from your shell
$ R

# install some packages if you don't have them on your system
> install.packages("PerformanceAnalytics")

# review our data
> df = read.csv("data/train.csv", header=T, nrow=100)
> head(df)
> summary(df)

> sapply(df, is.factor)
> cor(df[sapply(df, function(x) !is.factor(x))])
> plot(df[,10:20])

> library(PerformanceAnalytics)
> ndf=df[sapply(df, function(x) !is.factor(x))]
> chart.Correlation(ndf[,2:10], histogram=T)
```

### Our first models
Read and run `xgb_kernel.R` and `xgb_kernel.py` models.

To avoid overhead with static attributes let's create a new dataset where we'll pre-process the data and leave only attributes we're interested in:

```
./transform.py --fin=data/train.csv --fout=data/new_train.csv --target=QuoteConversion_Flag --drops=$drops
./transform.py --fin=data/test.csv --fout=data/new_test.csv --drops=$drops
# inspect output of our data reduction step
ls -alh data/
head -2 data/train.csv
head -2 data/new_train.csv
```

Now, we can re-read our data in R or Python and check how much memory we saved.
Answer: ipython (78MB), train dataset via pandas (608MB), reduced
train dataset via pandas (480MB).

##### Coding our model
Let's explore R and Python code in `xgb.R` and `xgb.py`, we'll concentrate on python version.

We can run it as following:
```
./xgb.py --ftrain=data/new_train.csv --ftest=data/new_test.csv --date=Original_Quote_Date --params=params.json --verbose
...
ROC AUC 0.9513833324610302
```
It provides as ROC AUC 0.95%. Can we improve this result?

One way to do it is to play with parameters, see `params.json`. For example, we
can increase number of training iterations from 10 to a 100 and observe how
our model will improve. Here the results:
```
[10]	eval-auc:0.952483	train-auc:0.953524
...
[99]	eval-auc:0.961165	train-auc:0.979542
('AUC', 0.9611651593829729)
```
The AUC score increased from 0.95% to 0.96%. Please note that ratio of
validation to train sample drops from 0.9989 to 0.9812. Usually it is good to
keep it close to 1 to avoid over-fitting. A good ratio (use-case dependant)
should be in a range between 0.8 and 1.0.

For simplicity (to save time) we'll continue using 10 rounds to demonstrate
other techniques.

One approach is to use stacking models, i.e. run multiple xgb.py versions with different seed

```
rm -rf results
./xgb.sh
grep AUC results/*.log
results/xgb_seed17.log:AUC 0.9539558840809022
results/xgb_seed17.log:ROC AUC 0.9539558840809022
results/xgb_seed5.log:AUC 0.9529006571058944
results/xgb_seed5.log:ROC AUC 0.9529006571058944
results/xgb_seed9.log:AUC 0.9525306573170437
results/xgb_seed9.log:ROC AUC 0.9525306573170437
```

We can combine these results and average the predictions, but how we can test that our combined results are an improvement.


Let's split dataset and place a holdout sample in order to demonstrate stacking and averaging
```
./mknfold.py --fin=data/new_train.csv --headers True
```
this will create `data/new_train.csv.train` and `data/new_train.csv.valid` samples.

Students Tasks: 
- re-run xgb.sh on `data/new_train.csv.train` sample and predict data for `data/new_train.csv.valid` one
- compare true labels and our predictions and see how we improve the model
- create stacking model

##### Embedding features
Let's demonstrate code for embedding features, have a look at `keras_embeddings.py`. Then we can run it as following:
```
# look-up code in utils.py module and study preprocessing function
# it provides all pieces to find out cardinality of our categorical variables
# explain cardinality concept

# keras_embeddings module runs simple Neural Network and generate weights of
# embedded matrix, we store these weights for each categorical variables
./keras_embeddings.py --ftrain=data/new_train.csv --ftest=data/new_test.csv --date=Original_Quote_Date --params=params.json --fout=weights --verbose

# now we re-run xgb model with produced embeddings weights
./xgb.py --ftrain=data/new_train.csv --ftest=data/new_test.csv --date=Original_Quote_Date --params=params.json --verbose --weights=weights
ROC AUC 0.952733694360817
```
We improved our AUC score from 0.9513833324610302 to 0.952733694360817, or just
by `0.13%`.

Here is a set of steps to improve our score:

- run `xgb.py` with larger number of epochs and smaller learning rate
- learn about xgboost parameters and tune them up
- train NN for embedded features with different NN parameters/learning rate, etc.
- create a stacking model's average

Repeat and re-iterate !!!
