#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : embeddings.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: example of fast.ai approach to create embedding matrix for a given dataset
https://github.com/fastai/fastai/blob/master/courses/dl1/lesson3-rossman.ipynb
"""

# system modules
import os
import sys
import argparse

# numpy modules
import numpy as np

# pandas modules
import pandas as pd

# sklearn modules
from sklearn.cross_validation import train_test_split
from sklearn.model_selection import ShuffleSplit

# fast.ai modules
import fastai
from fastai.structured import proc_df, add_datepart
from fastai.column_data import ColumnarModelData

class OptionParser():
    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--fin", action="store",
            dest="fin", default="", help="Input file")
        self.parser.add_argument("--target", action="store",
            dest="target", default="target", help="Name of target column")
        self.parser.add_argument("--date_col", action="store",
            dest="date_col", default="date", help="Name of date column")
        self.parser.add_argument("--cat_vars", action="store",
            dest="cat_vars", default="", help="categorical vairables")
        self.parser.add_argument("--fout", action="store",
            dest="fout", default="", help="Output file")
        self.parser.add_argument("--bsize", action="store",
            dest="bsize", default=128, help="Batch size, default 128")
        self.parser.add_argument("--verbose", action="store_true",
            dest="verbose", default=False, help="verbose output")

def inv_y(a):
    return np.exp(a)

def exp_rmspe(y_pred, targ):
    targ = inv_y(targ)
    pct_var = (targ - inv_y(y_pred))/targ
    return math.sqrt(pct_var**2).mean()

def categorical_variables(df):
    "Return list of categorical varilables in our dataframe"
    return [c for c in df.columns if df[c].dtype=='object']

def run(path, target, date_col, cat_vars, bsize, verbose):
    "Main function we run over our dataset"
    df, val_idx, y_pred, cat_vars = preprocess(path, target, date_col, cat_vars, verbose=verbose)
    embedding(path, df, val_idx, y_pred, cat_vars, bsize)

def preprocess(path, target, date_col, cat_vars=None, seed=123, split=0.3, verbose=False):
    "Pre-process functino which construct our dataset to run"
    x_train = pd.read_csv(path)

    # create or define categorical variables
    if not cat_vars:
       cat_vars = categorical_variables(x_train)
    if verbose:
        print("Categorical variables:", cat_vars)

    # loop over categorical values and convert its data type to ordered category one
    for v in cat_vars:
        x_train[v] = x_train[v].astype('category').cat.as_ordered()

    # use fast.ai proc_df which performs scaling
    df, y, nas, mapper = proc_df(x_train, target, do_scale=True)
    y_train = y
    # sometimes it is wise to use log(y) or log(y+1) function to deal with outliers
#     y_train = np.log(y)
#     y_train = np.log(y+1)

    # add date's variables
    add_datepart(df, date_col, drop=False)

    # create list of validation indexes (validation set)
    rs = ShuffleSplit(n_splits=3, test_size=split, random_state=seed)
    val_idx = []
    for train_idx, valid_idx in rs.split(df):
        for i in valid_idx:
            val_idx.append(i)

    return df, val_idx, y_train, cat_vars

def embedding(path, df, val_idx, y_pred, cat_vars, bs=128):
    """
    Embedding function which converts categorical varialbels of our dataframe
    into embedding matrix
    """
    # create columnar model
    PATH = 'data/new/' # location of our input file, I should replace this
    print(df.head())
    md = ColumnarModelData.from_data_frame(PATH, val_idx, df, y_pred, cat_flds=cat_vars, bs=bs)

    # creat embedding matrix
    cat_sz = [(c, len(joined_samp[c].cat.categories)+1) for c in car_vars]
    emb_szs = [(c, min(50, (c+1)/2)) for _,c in cat_sz]

    # get learner
    m = md.get_learner(emb_szs, len(df.columns)-len(cat_vars),
		       0.04, 1, [1000,500], [0.001,0.01], y_range=y_range)
    lr = 1e-3

    # find best learning rate
    m.lr_find()
    m.sched.plot(100)

    # fit the model
    m.fit(lr, 3, metrics=[exp_rmspe])

def main():
    "Main function"
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    run(opts.fin, opts.target, opts.date_col, opts.cat_vars, opts.bsize, opts.verbose)

if __name__ == '__main__':
    main()
