#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : keras_embeddings.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: 
    Code based on: https://github.com/entron/entity-embedding-rossmann
    and lecture:
    https://www.youtube.com/watch?v=V2h3IOBDvrA&feature=youtu.be&list=PLfYUBJiXbdtS2UQRzyrxmyVHoGW0gmLSM&t=6147
    and additional materials
    https://medium.com/@apiltamang/learning-entity-embeddings-in-one-breath-b35da807b596
"""

# system modules
import os
import sys
import time
import json
import pickle
import argparse

# numpy modules
import numpy as np

# pandas modules
import pandas as pd

# load keras modules
from keras.models import Sequential
from keras.models import Model as KerasModel
from keras.layers import Input, Dense, Activation, Reshape, Embedding
from keras.callbacks import EarlyStopping
from keras.layers.merge import Add, Concatenate

# local modules
from utils import preprocess, OptionParser

def cat_features(xdf, cat_vars):
    "Return categorical dataset"
    features = []
    for cat in cat_vars:
        features.append(xdf[cat])
    return features

def run(ftrain, fvalid, ftest, fout, date_col, params, verbose):
    "Main function we run over our dataset"
    print("Train file     : %s" % ftrain)
    print("Validation file: %s" % fvalid)
    print("Test file      : %s" % ftest)
    print("Date column    : %s" % date_col)
    print("Parameters     : %s" % json.dumps(params))

    # preprocess our data
    X_train, X_valid, X_test, y_train, y_valid, ids, cat_vars, cat_sz, emb_szs = \
            preprocess(ftrain, fvalid, ftest, date_col, params, verbose)

    # construct and fit the model
    embedding_model(fout, X_train, X_valid, X_test, y_train, y_valid, cat_vars, cat_sz, emb_szs, params, verbose)

def embedding_model(fout, X_train, X_valid, X_test, y_train, y_valid, cat_vars, cat_sz, emb_szs, params, verbose):
    """
    Embedding model train and converts categorical varialbels of our dataframe
    into embedding matrix
    """

    print("Cardinality of categorical variables")
    models = [] # list of models for embedded matrix
    inputs = [] # list of input models
    outputs = [] # list of output models
    for cat, emb in zip(cat_sz, emb_szs):
        if verbose:
            print(cat, emb)
        name = cat[0]
        input_dim = cat[1] # cardinality of categorical variable
        output_dim = emb[1] # embedded size of categorical variable

        # functional way to define NN layers
        input_var = Input(shape=(1,))
        output_var = Embedding(input_dim, output_dim, name=name)(input_var)
        output_var = Reshape(target_shape=(output_dim,))(output_var)
        inputs.append(input_var)
        outputs.append(output_var)

    # functional way to define NN model
    input_model = inputs
    output_model = Concatenate()(outputs)
    output_model = Dense(1000, kernel_initializer="uniform")(output_model)
    output_model = Activation('relu')(output_model)
    output_model = Dense(500, kernel_initializer="uniform")(output_model)
    output_model = Activation('relu')(output_model)
    output_model = Dense(1)(output_model)
    output_model = Activation('sigmoid')(output_model)
    model = KerasModel(inputs=input_model, outputs=output_model)

    # compile our neural net
    model.compile(loss='mean_absolute_error', optimizer='adam')

    # add callbacks
    early_stopping = EarlyStopping(monitor="val_loss", patience=5)

    # fit the model
    bsize = params.get('batch_size', 128)
    epochs = params.get('epochs', 50)
    level = 1 if verbose else 0
    model.fit(cat_features(X_train, cat_vars), y=y_train,
            batch_size=bsize, epochs=epochs, verbose=level,
            validation_data=(cat_features(X_valid, cat_vars), y_valid),
            callbacks=[early_stopping])

    # loop over model layers and save embeddings weights of categorical varilables
    # we write weights into fout area with a file name of categorical variable
    # e.g. "year" -> weights/year.pkl
    for layer in model.layers:
        conf = layer.get_config()
        name = conf['name']
        if name in cat_vars:
            if verbose:
                print("Layer  : %s" % conf)
                print("Weights: %s" % layer.get_weights())
            if fout:
                fname = '%s/%s.pkl' % (fout, name)
                if verbose:
                    print("Save : %s" % fname)
                with open(fname, 'wb') as ostream:
                    pickle.dump(layer.get_weights(), ostream)

    # make predictions
#     pred = model.predict(cat_features(X_valid, cat_vars)).flatten()
#     print("AUC", metrics.roc_auc_score(y_valid, pred))

    # create AUC/ROC plot
#     plot_roc_curve(y_valid, pred)

def main():
    "Main function"
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    time0 = time.time()
    params = json.load(open(opts.params)) if opts.params else {}
    run(opts.ftrain, opts.fvalid, opts.ftest, opts.fout, opts.date_col, params, opts.verbose)
    print("Done in %s sec" % (time.time()-time0))

if __name__ == '__main__':
    main()
