#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : merge_preds.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: merge predictions from input files based on given function
"""

# system modules
import os
import sys
import math
import gzip
import glob
import optparse

import numpy as np

class OptionParser():
    def __init__(self):
        "User based option parser"
        self.parser = optparse.OptionParser()
        self.parser.add_option("--fin", action="store", type="string",
            dest="fin", default="", help="Input file list use comma as separator")
        self.parser.add_option("--fout", action="store", type="string",
            dest="fout", default="", help="Output file")
        self.parser.add_option("--func", action="store", type="string",
            dest="func", default="mean",
            help="apply function, e.g. mean, median, hmean, gmean, default mean")
        self.parser.add_option("--weights", action="store", type="string",
            dest="weights", default="", help="weights for arithmetic mean")
        self.parser.add_option("--denominator", action="store", type="float",
            dest="den", default=0, help="denominator for arithmetic sum")
    def get_opt(self):
        "Return list of options"
        return self.parser.parse_args()

def fapply(name, points, weights=None, den=None):
    "Apply given function name to set of points"
    if name == 'wmean' or weights: # weighted mean
        den = sum(weights)
        nom = sum([w*v for v, w in zip(points, weights)])
        return nom/den
    elif  name == 'mean':
        return np.mean(points)
    elif name == 'median':
        return np.median(points)
    elif name == 'avg':
        tot = points[0]
        for idx in range(1, len(points)):
            tot += points[idx]
            tot = tot/2.
        return tot
    elif den: # common denominator
        tot = sum(points)
        return tot/den
    elif name == 'gmean': # geometric mean
        val = 1
        for item in points:
            val *= item
        return math.pow(val, 1./len(points))
    elif name == 'hmean': # harmonic mean
        data = []
        for item in points:
            if  not item:
                data.append(1e-9)
            else:
                data.append(item)
        tot = sum([1/p for p in data])
        return len(points)/tot
    else:
        raise NotImplemented

def merge(fin, fout, name='mean', weights=None, den=None):
    "Merge predictions from two different files"
    if  fin.find('*') != -1:
        files = glob.glob(fin)
    else:
        files = [os.path.expanduser(f) for f in fin.split(',')]
    fstreams = []
    for fname in files:
        statinfo = os.stat(fname)
        if  statinfo.st_size < 1000:
            continue # prevent empty files
        print fname
        if  fname.endswith('.gz'):
            fstreams.append(gzip.open(fname, 'r'))
        else:
            fstreams.append(open(fname, 'r'))
    headers = None
    if  fout.endswith('.gz'):
        ostream = gzip.open(os.path.expanduser(fout), 'wb')
    else:
        ostream = open(os.path.expanduser(fout), 'wb')
    while True:
        if  not headers:
            headers = [f.readline().replace('\n', '') for f in fstreams][0]
            ostream.write(headers+'\n')
        else:
            idict = {}
            for idx, item in enumerate(fstreams):
                vals = item.readline().replace('\n', '').split(',')
                idict[idx] = vals
            if  not vals[0]:
                break
            points = []
            uids = set()
            for idx in idict.keys():
                vals = idict[idx]
                uids.add(vals[0])
                for jdx, item in enumerate(vals[1:]):
                    if  len(points)>jdx:
                        points[jdx].append(float(item))
                    else:
                        points.append([float(item)])
            uids = list(uids)
            if  len(uids) != 1:
                raise Exception("Wrongs uids: %s" % uids)
            uid = uids[0]
            pred = []
            for vals in points:
                pred.append(fapply(name, vals, weights, den))
            ostream.write('%s,%s\n' % (uid, ','.join([str(v) for v in pred])))
    ostream.close()

def main():
    "Main function"
    optmgr  = OptionParser()
    opts, _ = optmgr.get_opt()
    if  opts.weights:
        weights = [float(w) for w in opts.weights.split(',')]
    else:
        weights = None
    merge(opts.fin, opts.fout, opts.func, weights, opts.den)

if __name__ == '__main__':
    main()
