#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : transform_csv.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Perform transformation with given CSV file. The allowed transformation are
drop multiple columns, convert dataframe into classification problem, rename given target
column.
"""
from __future__ import print_function

# system modules
import os
import re
import sys
import bz2
import gzip
from math import log
import argparse

# third-party modules
import pandas as pd

class OptionParser():
    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--fin", action="store",
            dest="fin", default="", help="Input file")
        self.parser.add_argument("--fout", action="store",
            dest="fout", default="", help="Output file")
        self.parser.add_argument("--idcol", action="store",
            dest="idcol", default="", help="id column name")
        self.parser.add_argument("--target", action="store",
            dest="target", default="", help="Target column name")
        self.parser.add_argument("--drops", action="store",
            dest="drops", default="", help="drops column names (comma separated)")
        self.parser.add_argument("--verbose", action="store_true",
            dest="verbose", default=0, help="Turn on verbose output")

class GzipFile(gzip.GzipFile):
    def __enter__(self):
        "Context manager enter method"
        if self.fileobj is None:
            raise ValueError("I/O operation on closed GzipFile object")
        return self

    def __exit__(self, *args):
        "Context manager exit method"
        self.close()

def fopen(fin, mode='r'):
    "Return file descriptor for given file"
    if  fin.endswith('.gz'):
        stream = gzip.open(fin, mode)
        # if we use old python we switch to custom gzip class to support
        # context manager and with statements
        if  not hasattr(stream, "__exit__"):
            stream = GzipFile(fin, mode)
    elif  fin.endswith('.bz2'):
        stream = bz2.BZ2File(fin, mode)
    else:
        stream = open(fin, mode)
    return stream

def transform(fin, fout, idcol, target, drops, verbose=0):
    "Perform transformation on given CSV file"
    istream = fopen(fin, 'r')
    ostream = fopen(fout, 'wb')
    headers = False
    count = 0
    for line in istream.readlines():
        if  not headers:
            headers = line.replace('\n', '').replace('"', '').split(',')
            if  drops:
                new_headers = ['ids']
                for idx, val in enumerate(headers):
                    if  val in drops or val == target or val == idcol:
                        continue
                    new_headers.append(val)
                if target:
                    ostream.write(','.join(new_headers)+',target\n')
                else:
                    ostream.write(','.join(new_headers)+'\n')
            print("headers", len(headers))
            continue
        line = line.replace('\n', '').replace('<nil>', '-1').replace('NA', '-1')
        # special treamtement for "1,123" which should be read as 1123
        vals = line.split('"')
        vals[3] = vals[3].replace(',','')
        line = '"'.join(vals)
        # proceed as normal
        vals = []
        for item in line.split(','):
            if item:
                vals.append(item.replace('"', ''))
            else:
                vals.append(-1)
        if  len(vals) != len(headers):
            raise Exception("Unable to parse line, #values != #headers", item, len(vals), len(headers))
        row = dict(zip(headers, vals))
        if  not idcol:
            row['ids'] = count
        new_vals = []
        for key in new_headers:
            if  key in drops or key == target or key == idcol:
                continue
            new_vals.append(str(row[key]))
        if target in row.keys():
            tval = row[target]
            new_vals.append(str(tval))
        ostream.write(','.join(new_vals)+'\n')
        count += 1
    istream.close()
    ostream.close()

def main():
    "Main function"
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    drops = opts.drops.split(',')
    transform(opts.fin, opts.fout, opts.idcol, opts.target, drops, opts.verbose)

if __name__ == '__main__':
    main()
