#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : utils.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Common set of utilities we use in xgb and keras_embeddings modules
"""

# system modules
import os
import sys
import json
import pickle
import argparse
from datetime import datetime

# numpy modules
import numpy as np

# pandas modules
import pandas as pd

# sklearn modules
from sklearn.cross_validation import train_test_split
from sklearn.preprocessing import Normalizer, StandardScaler, LabelEncoder
from sklearn import metrics

# matplotlib
import matplotlib.pyplot as plt

class OptionParser():
    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--ftrain", action="store",
            dest="ftrain", default="", help="Input train file, default None")
        self.parser.add_argument("--fvalid", action="store",
            dest="fvalid", default="", help="Input validation file, default None")
        self.parser.add_argument("--ftest", action="store",
            dest="ftest", default="", help="Input test file, default None")
        self.parser.add_argument("--date_col", action="store",
            dest="date_col", default="date", help="Name of date column, default date")
        self.parser.add_argument("--fout", action="store",
            dest="fout", default="", help="Output file")
        self.parser.add_argument("--params", action="store",
            dest="params", default="", help="parameters file")
        self.parser.add_argument("--weights", action="store",
            dest="weights", default="", help="Location of embeddings weights")
        self.parser.add_argument("--verbose", action="store_true",
            dest="verbose", default=False, help="verbose output")

def func_year(x):
    dt = datetime.strptime(x, '%Y-%m-%d')
    return str(dt.year) # make it categorical

def func_month(x):
    dt = datetime.strptime(x, '%Y-%m-%d')
    return str(dt.month) # make it categorical

def func_day(x):
    dt = datetime.strptime(x, '%Y-%m-%d')
    return str(dt.day) # make it categorical

def add_date_attributes(xdf, date_col):
    "Add date attributes, e.g. year, month, day to given data frame"
    xdf['year'] = xdf[date_col].apply(func_year)
    xdf['month'] = xdf[date_col].apply(func_month)
    xdf['day'] = xdf[date_col].apply(func_day)
    return xdf

def categorical_variables(df):
    "Return list of categorical varilables in our dataframe"
    return [c for c in df.columns if df[c].dtype=='object']

def one_hot_encoding(X_train, X_valid, X_test, cat_vars):
    "Perform one-hot-encoding"
    for col in X_train.columns:
        if col in cat_vars:
            lbl = LabelEncoder()
            lbl.fit(list(X_train[col].values)+list(X_valid[col].values)+list(X_test[col].values))
            X_train[col] = lbl.transform(list(X_train[col].values))
            X_valid[col] = lbl.transform(list(X_valid[col].values))
            X_test[col] = lbl.transform(list(X_test[col].values))
    return X_train, X_valid, X_test

def plot_roc_curve(y_valid, pred):
    fpr, tpr, thresholds = metrics.roc_curve(y_valid, pred)
    roc_auc = metrics.auc(fpr, tpr)
    print("valid", y_valid)
    print("preds", pred)
    print("ROC AUC", roc_auc)
    print("FPR", fpr)
    print("TPR", tpr)

    plt.figure()
    lw = 2
    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic')
    plt.legend(loc="lower right")
    plt.savefig('roc.pdf')

def preprocess(ftrain, fvalid, ftest, date_col, params, verbose=False):
    "Build and fit model for given train/validation/test/out files"
    seed = params.get('seed', 12345)
    split = params.get('split', 0.3)
    np.random.seed(seed)

    # we either read validation file or make a train/validation split
    if  fvalid:
        X_train = pd.read_csv(ftrain)
        X_valid = pd.read_csv(fvalid)

        y_train = X_train.target.values
        X_train = X_train.drop('target', axis=1)
        y_valid = X_valid.target.values
        X_valid = X_valid.drop('target', axis=1)
    else:
        print("No validation is provided will split %s" % split)
        X_train = pd.read_csv(ftrain)
        y_train = X_train.target.values
        X_train = X_train.drop('target', axis=1)
        columns = X_train.columns
        X_train, X_valid, y_train, y_valid = \
                train_test_split(X_train, y_train, random_state=seed, test_size=split)
        # convert X_train/valid back to DataFrame since split function returns ndarray
        X_train = pd.DataFrame(X_train)
        X_train.columns = columns
        X_valid = pd.DataFrame(X_valid)
        X_valid.columns = columns

    # read test file and extract ids
    X_test = pd.read_csv(ftest)
    ids = X_test.ids.values

    X_train = add_date_attributes(X_train, date_col)
    X_valid = add_date_attributes(X_valid, date_col)
    X_test = add_date_attributes(X_test, date_col)

    # drop ids since we don't need them anymore
    for col in ['ids', date_col]:
        X_train = X_train.drop(col, axis=1)
        X_valid = X_valid.drop(col, axis=1)
        X_test = X_test.drop(col, axis=1)

    # fill out NAs
    X_train = X_train.fillna(-1)
    X_valid = X_valid.fillna(-1)
    X_test = X_test.fillna(-1)

    if verbose:
        print("train columns", len(list(X_train.columns)))
        print("valid columns", len(list(X_valid.columns)))
        print("test  columns", len(list(X_test.columns)))

    # create full dataframe which will hold all our data
    X_full = X_train.append(X_valid).append(X_test)

    # create or define categorical variables
    cat_vars = categorical_variables(X_train)
    for name in cat_vars:
        X_train[name] = X_train[name].astype('category').cat.as_ordered()
        X_valid[name] = X_valid[name].astype('category').cat.as_ordered()
        X_test[name] = X_test[name].astype('category').cat.as_ordered()
        X_full[name] = X_full[name].astype('category').cat.as_ordered()

    if verbose:
        print("Categorical variables:", len(cat_vars))

    # max size of embedded matrix
    max_emb = params.get('max_embeddings_size', 50)

    # find categorical sizes from full dataset (train+valid+test)
    cat_sz = [(c, len(X_full[c].cat.categories)+1) for c in cat_vars]

    # embedded sizes
    emb_szs = [(c, min(max_emb, (c+1)//2)) for _, c in cat_sz]

    # perform one-hot-encoding
    X_train, X_valid, X_test = one_hot_encoding(X_train, X_valid, X_test, cat_vars)

    # add aggregated columns, e.g. sum of all attributes
    X_train["sum"] = X_train.sum(axis=1)
    X_valid["sum"] = X_valid.sum(axis=1)
    X_test["sum"]  = X_test.sum(axis=1)

    if verbose:
        name = cat_vars[0]
        print("cat variable", name, "after one-hot-encoding")
        print(X_train[name].head())

    return X_train, X_valid, X_test, y_train, y_valid, ids, cat_vars, cat_sz, emb_szs
