#!/usr/bin/env Rscript

library(xgboost)
library(bit64)
library(methods)
library(parallel)
library(stringr)
library(data.table)
library(pROC)

ncores = detectCores()

set.seed(1234)

# options will be available via opts dataframe
source("optsparser.R")

# helper function to drop columns from given dataset
drop <- function(xdf, drops) {
    return(xdf[,!names(xdf) %in% drops])
}

myfunc <- function(x) log(x+2)

print(sprintf("Read %s", opts$train))
train = as.data.frame(fread(sprintf("%s", opts$train), header=TRUE))
train[is.na(train)] = missing
print(sprintf("Read %s", opts$test))
test = as.data.frame(fread(sprintf("%s", opts$test), header=TRUE))
test[is.na(test)] = missing
test.ids = test$ids
test$ids = NULL

# convert categorical values into numerical
cat("assuming text variables are categorical & replacing them with numeric ids\n")
feature.names <- names(train)
for (f in feature.names) {
  if (class(train[[f]])=="character") {
    levels <- unique(c(train[[f]], test[[f]]))
    train[[f]] <- as.integer(factor(train[[f]], levels=levels))
    test[[f]]  <- as.integer(factor(test[[f]],  levels=levels))
  }
}
missing = 0.0

# drops
drops = c("GeographicField10A", "GeographicField10B", "PropertyField6", "Year", "Month")
train = drop(train, drops)
test = drop(test, drops)

print(sprintf("Train nrow=%s, ncol=%s", nrow(train), ncol(train)))
print(sprintf("Test  nrow=%s, ncol=%s", nrow(test), ncol(test)))
 
# transformation
#train = lapply(train, myfunc)
#test = lapply(test, myfunc)

# split train sample into train and validation
smp_size <- floor(0.75 * nrow(train))
train_ind <- sample(seq_len(nrow(train)), size = smp_size)
x.train <- train[train_ind, ]
x.valid <- train[-train_ind, ]
y.train <- x.train$target
y.valid <- x.valid$target

x.train$target <- NULL
x.valid$target <- NULL

print(sprintf("### y.train=%s", length(y.train)))
print(sprintf("### y.valid=%s", length(y.valid)))

# scale data
#range01 <- function(x){(x-min(x))/(max(x)-min(x))}
#y = range01(cls.target)


# construct xgb.DMatrix, i.e. prepare data for training/prediction
print(sprintf("xgb.train nrow=%s, ncol=%s", nrow(x.train), ncol(x.train)))
x = as.matrix(x.train)
x = matrix(as.numeric(x), nrow(x), ncol(x))
x[is.na(x)] = missing
xgb.train = xgb.DMatrix(data=x, label=y.train)

print(sprintf("xgb.valid nrow=%s, ncol=%s", nrow(x.valid), ncol(x.valid)))
x = as.matrix(x.valid)
x = matrix(as.numeric(x), nrow(x), ncol(x))
x[is.na(x)] = missing
xgb.valid = xgb.DMatrix(data=x, label=y.valid)

watchlist<-list(val=xgb.valid, train=xgb.train)
param <- list(  objective           = "binary:logistic", 
                booster = "gbtree",
                eval_metric = "auc",
                eta                 = 0.01, # 0.02, # 0.06, #0.01,
                max_depth           = 15, #7, #changed from default of 8
                subsample           = 0.8, #0.82, # 0.7
                colsample_bytree    = 0.8 #0.66 # 0.7
                #num_parallel_tree   = 2
                # alpha = 0.0001, 
                # lambda = 1
)

model <- xgb.train( params              = param, 
                    data                = xgb.train, 
                    nrounds             = opts$nround, #1800, 
                    verbose             = 0,  #1
                    #early.stop.round    = 150,
                    watchlist           = watchlist,
                    maximize            = TRUE
)

sub <- data.frame(QuoteNumber=seq(1, nrow(x.valid), 1))
sub$QuoteConversion_Flag <- NA
# Make prediction for valid set
x = as.matrix(x.valid)
x = matrix(as.numeric(x), nrow(x), ncol(x))
x[is.na(x)] = missing
pred = predict(model, x, missing=missing)
print("Prediction")
print(head(pred))
print("AUC")
print(auc(y.valid, pred))
rocobj <- roc(y.valid, pred)
pdf("xgbR.pdf")
plot(rocobj)
dev.off()

sub$QuoteConversion_Flag = pred
write.csv(sub, sprintf('%s', opts$out), row.names=FALSE, quote=FALSE)

sub <- data.frame(QuoteNumber=test.ids)
sub$QuoteConversion_Flag <- NA
dtest = data.matrix(test)
dtest[is.na(dtest)] = missing
pred = predict(model, xgb.DMatrix(data=dtest), missing=missing)
sub$QuoteConversion_Flag = pred
write.csv(sub,"xgb_R.csv",row.names=FALSE,quote=FALSE)
