#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : xgb.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: 
"""

# system modules
import os
import sys
import time
import json
import pprint
import pickle

# pandas modules
import pandas as pd

# numpy modules
import numpy as np

# xgboost modules
import xgboost as xgb

# sklearn modules
from sklearn import metrics

# matplotlib
import matplotlib.pyplot as plt

# local modules
from utils import preprocess, OptionParser, plot_roc_curve

def lookup(x, w):
    "Helper function to look-up embeddings values from weights matrix"
    return w[x]

def merge_embeddings(xdf, cat, emb, w, verbose):
    """
    Helper function to merge embeddings values into given dataframe
    """
    if verbose:
        print("embeddings category: '%s' embeddings shape %s" % (cat, np.shape(w)))
    if cat not in list(xdf.columns):
        print("categorical variable '%s' not found in data-frame" % cat)
        print(type(xdf), np.shape(xdf), list(xdf.columns))
        sys.exit(1)

    # perform look-up in our weights matrix for given value of categorical variables
    emd_values = list(xdf[cat].apply(lookup, args=(w,)))

    # create vector of names for embeddings columns
    emd_names = ['%s_%s' % (cat, i) for i in range(len(w[0]))]

    if verbose:
        print("new embeddings columns", len(emd_names), "with shape", np.shape(emd_values))
    emd_df = pd.DataFrame(emd_values, columns=emd_names)
    if verbose:
        print("xdf dim", np.shape(xdf), "emb dim", np.shape(emd_df))

    # merge emdedded df with our data-frame column-wise, to stick two dataframes
    # together we reset their indexes
    names = list(xdf.columns) + emd_names
    xdf.reset_index(drop=True, inplace=True)
    emd_df.reset_index(drop=True, inplace=True)
    xdf = pd.concat([xdf, emd_df], axis=1, ignore_index=True)
    xdf.columns = names

    # drop categorical variable from our data-frame since we replace it with
    # emdedded values
    if cat in list(xdf.columns):
        xdf = xdf.drop(cat, axis=1)
    if verbose:
        print("new dimension of dataframe:", np.shape(xdf), type(xdf))
    return xdf

def add_weights(X_train, X_valid, X_test, weights, emb, verbose=False):
    "Add embedding weights to our dataframes"
    if not weights:
        return X_train, X_valid, X_test
    for name in os.listdir(weights):
        if name.endswith('pkl'):
            fname = '%s/%s' % (weights, name)
            with open(fname, 'rb') as istream:
                # load weight matrix, it is in form of [array([[], [])]
                w = pickle.load(istream)[0]
                if verbose:
                    print("loading", fname)
                # name of categorical variable
                cat = name.split('.')[0]
                X_train = merge_embeddings(X_train, cat, emb, w, verbose)
                X_valid = merge_embeddings(X_valid, cat, emb, w, verbose)
                X_test = merge_embeddings(X_test, cat, emb, w, verbose)
    return X_train, X_valid, X_test

def model(fout, X_train, X_valid, X_test, y_train, y_valid, ids, cat_vars, cat_sz, emb_szs, params, verbose):
    "Build and fit model for given train/validation/test/out files"

    nround = params.get('nround', 10)
    early_stopping_rounds = params.get('early_stopping_rounds', 10)

    print("Train shape", np.shape(X_train))
    print("Valid shape", np.shape(X_valid))
    print("Test  shape", np.shape(X_test))

    # preparre DMatrix object for training/fitting
    dtrain = xgb.DMatrix(X_train, label=y_train)
    deval = xgb.DMatrix(X_valid, label=y_valid)
    dtest = xgb.DMatrix(X_test)

    # model parameters
    args = {'bst:max_depth':7,
            'bst:eta':0.02,
            'bst:subsample':0.86,
            'bst:colsample_bytree': 0.68,
            'eval_metric': 'auc',
            'silent':0,
            'objective':'binary:logistic'}
    if verbose:
        print("model parameters")
        pprint.pprint(args)

    # use evaluation list while traning
    evallist  = [(deval,'eval'), (dtrain,'train')]
    # train our model with early stopping that we'll see that we don't overfit
    bst = xgb.train(args, dtrain, nround, evallist, early_stopping_rounds=early_stopping_rounds)

    # validate results
    pred = bst.predict(deval)
    print("AUC", metrics.roc_auc_score(y_valid, pred))

    # create AUC/ROC plot
    plot_roc_curve(y_valid, pred)

    # make prediction
    if fout:
        pred = bst.predict(dtest)
        data = {'QuoteNumber':ids, 'QuoteConversion_Flag': pred}
        sub = pd.DataFrame(data, columns=['QuoteNumber', 'QuoteConversion_Flag'])
        print("Write prediction to %s" % fout)
        sub.to_csv(fout, index=False)

def run(ftrain, fvalid, ftest, fout, date_col, params, weights, verbose):
    "Main function we run over our dataset"
    print("Train file     : %s" % ftrain)
    print("Validation file: %s" % fvalid)
    print("Test file      : %s" % ftest)
    print("Date column    : %s" % date_col)
    print("Parameters     : %s" % json.dumps(params))

    # preprocess our data
    X_train, X_valid, X_test, y_train, y_valid, ids, cat_vars, cat_sz, emb_szs = \
            preprocess(ftrain, fvalid, ftest, date_col, params, verbose)

    if verbose:
        print("Categorical cardinality")
        print(cat_sz)
        print("Embedding matrix")
        print(emb_szs)

    # if we provides embedding weights we'll add them to our data frames
    if weights:
        X_train, X_valid, X_test = add_weights(X_train, X_valid, X_test, weights, cat_sz, verbose)

    # construct and fit the model
    model(fout, X_train, X_valid, X_test, y_train, y_valid, ids, cat_vars, cat_sz, emb_szs, params, verbose)

def main():
    "Main function"
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    time0 = time.time()
    params = json.load(open(opts.params)) if opts.params else {}
    run(opts.ftrain, opts.fvalid, opts.ftest, opts.fout, opts.date_col, params, opts.weights, opts.verbose)
    print("Done in %s sec" % (time.time()-time0))

if __name__ == '__main__':
    main()

